/*
 * Este ejemplo sirve para avisar del tipo de tarifa eléctrica que corresponde al día y
 * a la hora actual. Se adjunta el calendario de Iberdrola con las diferencias entre días
 * y los horarios de los cuatro tipos de dia
 * TIPOS DE DIA--> BAJO, MEDIO, ALTO y PICO
 * TARIFAS-------> VALLE, LLANO, PUNTA 100% y PUNTA 300%
 */

#include <Wire.h>
#include <RTC_lib.h>

#include <LCD.h>
#include <LiquidCrystal_I2C.h>
#define LCD_DIR 0x27

#include <SD.h>
File  Archivo;
File Registro;

#define INVIERNO false
#define VERANO true
//Asignación numérica de las tarifas y tipos de día
#define VALLE 1
#define LLANO 2
#define PUNTA100 3
#define PUNTA300 4
#define BAJO 1
#define MEDIO 2
#define ALTO 3
#define PICO 4

LiquidCrystal_I2C lcd(LCD_DIR, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);

RTC rtc;
Reloj y;

//Variable para activar y desactivar el control automático
volatile bool manual = false;
//Se declara una varible volatil que será la que modifique la interrupción
volatile bool cambio = false;
//Variable época, que indica si es verano o invierno, necesaria para los días tipo ALTO
 bool epoca = VERANO;
//Vector que almacena los días festivos, el tamaño es 24, 12 para los días y 12 para el mes
int festivos[24];
//La variable estado sirve para seguir una secuencia con las tarifas correctas cada nuevo día
 int  estado = 0;
//La variable tarifadia indica que tipo de día es, se inicialia a BAJO y varía cuando estado es 1
 int  tarifadia = BAJO;
//La tarifa se inicialia en VALLE y se actualiza según el estado
 int  tarifa = VALLE;

 //Horarios del suministro
int h_ini=0, m_ini=0, h_fin=0, m_fin=0;
//Variable que indica si el suministro esta activo o no
bool suministro = false;
 
//Función para comprobar que día es y asignar una tarifa
 int tarifaDia();
//Función que lee los datos almacenados en el archivo
 void compruebaArchivo();
//Función que recorre el vector de días festivos
 bool recorre_festivos(int dia, int mes);
  

void setup()
{
   Serial.begin(9600);
  while(!Serial); 

//Se lee el fichero de texto donde se indican los días festivos y los horarios de suministro automático
  compruebaArchivo();

//Inicialización del LCD
  lcd.begin (16,2);
  lcd.setBacklight(ON);
  lcd.clear();
  
//Una primera lectura de los valores de los registros
  y=rtc.leerHora();
//Se establece el modo 24h
  rtc.Modo12h(OFF);
  
  rtc.limpiaAlarma();
  
  //La alarma 1 de tipo 4 irá variando, y depende de la hora / min / seg
  rtc.alarmasON(4);
  //Se inicializa la alarma a las 00:00, para que comience en un cambio de día y así evitar
  //iniciar la secuencia en un estado más avanzado o atrasado al de la hora que le corresponde
  rtc.escribeAlarma(0,0,0,0);
  //Activamos interrupciones en el pin 2 de Arduino cuando haya un flanco de bajada
  attachInterrupt(0,interrup,FALLING); 

  //Se habilita la segunda alarma de tipo 6, depende de la hora / min
  //Con ella se controlarán los horarios de activación del suministro
  rtc.alarmasON(6);
  rtc.escribeAlarma(0,h_ini,m_ini,0);

  //Activamos interrupciones en el pin 3 de Arduino cuando haya un flanco de bajada
  //Con esta interrupcion se cambiara de modo manual a automático
  attachInterrupt(1,autom,FALLING); 

}

void loop()
{
  if(estado==1)
  {
    //Siempre que llegue al estado 1 es un cambio de día, por lo que leemos los registros
    y = rtc.leerHora();
    
    //La tarifa siempre es Valle por la noche, de 00:00 a 8:00 o 10:00
    tarifa = VALLE;
    
    //Determinar el tipo de día llamando a la funcion tarifaDia(), Bajo, Medio, Alto o Pico
    tarifadia = tarifaDia();
    switch(tarifadia)
    {
      case BAJO:
        estado = 0;
        //escribimos la proxima alarma, que indicará el cambio de tarifa
        rtc.escribeAlarma(0,0,0,0);
      break;
        
      case MEDIO:
        estado = 2;
        //escribimos la proxima alarma, que indicará el cambio de tarifa
        rtc.escribeAlarma(0,10,0,0);
      break;
    
      case ALTO:
        estado = 4;
        //escribimos la proxima alarma, que indicará el cambio de tarifa
        rtc.escribeAlarma(0,8,0,0);        
      break;

      case PICO:
        estado = 7;
        //escribimos la proxima alarma, que indicará el cambio de tarifa
        rtc.escribeAlarma(0,8,0,0);        
      break; 
    }

      //mostrar por LCD
      lcd.clear();
      lcd.setCursor(0,0);
      lcd.print("Es dia "); lcd.print(y.dia); lcd.print(",mes ");lcd.print(y.mes);
      lcd.setCursor(0,1);
      lcd.print("Tipo:"); lcd.print(tarifadia); lcd.print(" Tarifa:");lcd.print(tarifa);
      //Se escribe en un fichero de texto llamado registro el cambio de día
      Registro = SD.open("registro.txt", FILE_WRITE);
       if (Registro)
       {
          Registro.print("Es dia "); Registro.print(y.dia); Registro.print(", mes ");Registro.println(y.mes);
          Registro.print("El tipo de dia correspondiente es: ");Registro.println(tarifadia);
          Registro.close();
       }
  }

//Se comprueba si está activado el modo manual o no
else if(manual==false)
{
//En caso de producirse una interrupción y cambio pase a ser TRUE
  if(cambio)
  {
//Se comprueba si la alarma activada es la alarma 1  
    if((rtc.leeRegistro(STATUS_REG)&&0b00000011)==0b00000001)
    {
 //Lectura del registro, por si hubiera variado (no debería)
    y = rtc.leerHora();
    //Se activa la alarma 1
    rtc.alarmasON(4);
     rtc.limpiaAlarma();
     
    //Dependiendo el estado en que se esté, se actualiza la tarifa y la proxima alarma
    switch(estado)
    {
      case 0:
        //Este estado va asociado siempre a la alarma de las 00:00, por lo tanto entra siempre que el día cambia
        estado = 1;
       break;

      case 2:
        tarifa = LLANO;
        rtc.escribeAlarma(0,22,0,0);
        estado = 3;
       break;

      case 3:
        tarifa = VALLE;
        rtc.escribeAlarma(0,00,0,0);
        estado = 0;
       break;

//Los siguientes 3 estados corresponden a los días tipo ALTO, ya que se producen 4 alarmas, la primera en el loop y las 3 siguientes

      case 4:
        tarifa = LLANO;
        if(epoca)
          rtc.escribeAlarma(0,11,0,0);
        else 
          rtc.escribeAlarma(0,18,0,0);
        estado = 5;
       break;

      case 5:
        tarifa = PUNTA100;
        if(epoca)
          rtc.escribeAlarma(0,15,0,0);
        else
          rtc.escribeAlarma(0,22,0,0);
        estado = 6;

       break;

      case 6:
        tarifa = LLANO;
        rtc.escribeAlarma(0,00,0,0);
        estado = 0;
       break;

//Los siguientes estados corresponden a dias tipo PICO

      case 7:
        tarifa = LLANO;
        rtc.escribeAlarma(0,9,0,0);
        estado = 8;
       break;

      case 8:
        tarifa = PUNTA300;
        rtc.escribeAlarma(0,14,0,0);
        estado = 9;

       break;

      case 9:
        tarifa = LLANO;
        rtc.escribeAlarma(0,17,0,0);
        estado = 10;

       break;

      case 10:
        tarifa = PUNTA300;
        rtc.escribeAlarma(0,22,0,0);
        estado = 6;

       break;
    }
    if(estado!=1)
    {
      //Mostrar por la pantalla del LCD
      //lcd.clear();
      lcd.setCursor(0,1);
      lcd.print("Tipo: "); lcd.print(tarifadia); lcd.print(" Tarifa:");lcd.print(tarifa);
      Registro = SD.open("registro.txt", FILE_WRITE);
      if (Registro)
      {
      Registro.print("Tarifa: ");Registro.println(tarifa);
      }
    }
   }
   //En caso de ser la alarma 2, se cambia el estado de suministro y se escribe la alarma de inicio o parada
    else
    {
      suministro = !suministro;
      //Se activa la alarma 2
      rtc.alarmasON(6);
      if(suministro==ON)
      {
        rtc.limpiaAlarma();
        rtc.escribeAlarma(0,h_fin,m_fin,0);
         Registro = SD.open("registro.txt", FILE_WRITE);
         if (Registro)
         {
            Registro.print("--La tarifa actual es ");Registro.print(tarifa);Registro.println("--");
            Registro.print("Se ha iniciado el suministro a las\t");
            Registro.print(h_ini);Registro.print(":");Registro.println(m_ini);
            Registro.close();
         }
        }
        else if(suministro==OFF)
       {
         Registro = SD.open("registro.txt", FILE_WRITE);
         if (Registro)
         {
            rtc.limpiaAlarma();
            rtc.escribeAlarma(0,h_ini,m_ini,0);
            Registro.print("Se ha cerrado el suministro a las\t");
            Registro.print(h_fin);Registro.print(":");Registro.println(m_fin);
            Registro.close();
         }
        }
      }
   //Se vuelve a poner cambio a FALSE, para poder variarla de nuevo en una nueva interrupción
  cambio=false;
  }
 }
}


//Esta función comprueba los valores que contiene la variable tipo Reloj y
//los cuales corresponderán a los de los registros, ya que interesa dia, diaSem y mes
//no hay problema por que lo lea algo más tarde desde que cambió en el RTC
int tarifaDia()
{
  int dia=1;
  //Los dias tipo BAJO son, del 28 de Abril al 24 de Mayo + del 4 al 22 de Agosto + Todos los Domingos (Incluidos ya en otra alarma)
  //Sabados de Marzo, Abril, Mayo, Junio, Julio, Agosto, Septiembre, Octubre y Noviembre + los festivos   
  if(y.diaSem==DOMINGO || (y.mes==4 && y.dia>27) || (y.mes==5 && y.dia<25) || (y.mes==8 && y.dia>3 && y.dia<24) || (y.diaSem==SABADO && y.mes>2 && y.mes<12) || recorre_festivos(y.dia,y.mes))  
      dia = BAJO;
     
  //Los dias tipo MEDIO son, el 31 de Marzo, el resto de Abril y Mayo que no son BAJOS, las dos primeras semanas de Junio, la ultima semana de agosto
  //del 15 de Septiembre al 24 de Octubre y el 8 de Diciembre
    else if((y.mes==3 && y.dia==31) || (y.mes==4) || (y.mes==5) || (y.mes==6 && y.dia<16) || (y.mes==8 && y.dia>24) || (y.mes==9 && y.dia>14) || (y.mes==10 && y.dia<25) || (y.mes==12 && y.dia==8))
             dia = MEDIO;
          
  //Los dias tipo ALTO son, el resto de Marzo, Junio, Julio, Agosto, Septiembre y Octubre que no es BAJO o MEDIO + hasta el 15 de Noviembre + 5 y 26 de Diciembre
        else if(y.mes==3 || y.mes==6 || y.mes==7 || y.mes==8 || y.mes==9 || y.mes==10 ||(y.mes==11 && y.dia<16) || (y.mes==11 && (y.dia==5||y.dia==26)) )           
              {
                //El valor de la variable cambioHorario es true si es verano y false si es invierno
                epoca = y.cambioHorario;
                dia = ALTO;
              }
  //El resto son dias de PICO, Enero, Febrero, segunda mitad de Noviembre y Diciembre
            else 
                dia = PICO;
        
  return dia;
}

void compruebaArchivo()
{
  char linea[10];
  char buffer[3];
    //Se muestra por pantalla que se va a iniciar la comunicación con la SD
  Serial.println("Comenzando la comunicación con la tarjeta SD");

  //Se establece como salida el pin correspondiente a SS.
  pinMode(10, OUTPUT);

  //Se muestra por el monitor si la comunicación se ha establecido correctamente
  //o ha habido algún tipo de error.
  if (!SD.begin(10)) {

    Serial.println("Se ha producido un fallo al iniciar la comunicación");
    return;
  }
  Serial.println("Se ha iniciado la comunicación correctamente");

   //Se abre el fichero para leer los datos escritos.
  Archivo = SD.open("datos.txt");

  //Si el archivo se ha abierto correctamente se almacenan los datos.
  if (Archivo) {

      //Se salta la primera línea, donde se indica como añadir los días festivos
      Archivo.readBytesUntil('\r', linea, sizeof(linea));
      Archivo.seek(Archivo.position() + 1); 
      //Se almacena el valor de cada día y mes festivo
      for(int i=0;i<sizeof(festivos)/2;++i)
      {
        buffer[0,1]='\0';
        Archivo.readBytesUntil('/', buffer, sizeof(buffer));
        festivos[i]=atol(buffer);
        buffer[0,1]='\0';
        Archivo.readBytesUntil('\r', buffer, sizeof(buffer));
        festivos[i+1]=atol(buffer);
        Archivo.seek(Archivo.position() + 1); 
      }
      buffer[0,1]='\0';
      Archivo.readBytesUntil('\r', linea, sizeof(linea));
      Archivo.seek(Archivo.position() + 1); 
      Archivo.readBytesUntil(':', buffer, sizeof(buffer));
      h_ini=atol(buffer);
      buffer[0,1]='\0';
      Archivo.readBytesUntil('\r', buffer, sizeof(buffer));
      m_ini=atol(buffer);
      Archivo.readBytesUntil('\r', linea, sizeof(linea));
      Archivo.seek(Archivo.position() + 1); 
      buffer[0,1]='\0';
      Archivo.readBytesUntil(':', buffer, sizeof(buffer));
      h_fin=atol(buffer);
      buffer[0,1]='\0';
      Archivo.readBytesUntil('\r', buffer, sizeof(buffer));
      m_fin=atol(buffer);
    //Si todo ha ido bien cierra el archivo para no perder datos.
    Archivo.close();
  }
    //En caso de que haya habido problemas abriendo datos.txt, se muestra por pantalla.
  else {

    Serial.println("El archivo datos.txt no se abrió correctamente");
  }
}

bool recorre_festivos(int dia, int mes)
{
  for(int i=0; i<sizeof(festivos)/2; ++i)
  {
    if(dia==festivos[i] && mes==festivos[i+1])
        return ON;
    ++i;
  }
  return OFF;
}

void interrup()
{
  cambio=true;
}

void autom()
{
  manual =!manual;
}

